/** 
 * \file 3dvector.hpp
 *  
 * \brief This file contains the  definition of a class for vectors in
 * 3D.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef t3DVECTOR_HPP
#define t3DVECTOR_HPP

#include <cmath>                // sqrt()
#include <sstream>              // std::stringstream

#include "mathmacros.hpp"       // is_equal(), is_zero(), sqr(), INVERSE()
#include "exceptionobject.hpp"  // ExceptionObject 


/** 
 * \defgroup CommonNameSpace Namespace common.
 * @{
 */

/**
 * \namespace common
 *
 * \brief The  namespace common contains the definition  of macros and
 * several  utility  classes, such  as  a  class  for representing  3D
 * vectors and  a class  for dealing with  exceptions in  a customized
 * way.
 */

namespace common {


  /**
   * \class t3DVector
   *
   * \brief This class represents a vector in 3D space.
   */
  class t3DVector {
  public:
    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /** 
     * \fn t3DVector()
     *
     * \brief Creates an instance of this class.
     */
    t3DVector() :
      _x( 0 ) ,
      _y( 0 ) , 
      _z( 0 ) 
    {
    }


    /** 
     * \fn t3DVector( const t3DVector& v )
     *
     * \brief Creates an instance  of this class from another instance
     * of this class.
     *
     * \param v A given 3D vector to be cloned.
     */
    t3DVector( const t3DVector& v ) : 
      _x( v._x ) , 
      _y( v._y ) , 
      _z( v._z ) 
    {
    }


    /** 
     * \fn t3DVector( double x , double y , double z )
     *
     * \brief  Creates an instance  of this  class and  initializes it
     * with a given set of coordinates.
     *
     * \param x The X coordinate of the vector.
     * \param y The Y coordinate of the vector.
     * \param z The Z coordinate of the vector.
     */
    t3DVector( 
	      double x ,
	      double y ,
	      double z 
	     )
    {
      this->set_coordinates( x , y , z ) ;
    }


    /** 
     * \fn void set_coordinates( const t3DVector& v )
     *
     * \brief  Changes the  coordinates of  this vector  so  that they
     * become equal to the respective coordinates of a given vector.
     *
     * \param v A vector in 3D space.
     */
    void set_coordinates( const t3DVector& v )
    {
      _x = v._x ;
      _y = v._y ;
      _z = v._z ;
    }


    /** 
     * \fn void set_coordinates( double x , double y , double z )
     *
     * \brief  Changes the  coordinates of  this vector.
     *
     * \param x The new value of the X coordinate.
     * \param y The new value of the Y coordinate.
     * \param z The new value of the Z coordinate.
     */
    void set_coordinates( double x , double y , double  z )
    {
      _x = x ;
      _y = y ;
      _z = z ;
    }


    /** 
     * \fn bool operator==( const t3DVector& v ) const
     *
     * \brief Verifies whether this vector is equal to a given vector.
     *
     * \return  The logic  value  true  if this  vector  has the  same
     * coordinates of a given vector. Otherwise, the logic value false
     * is returned.
     */
    bool operator==( const t3DVector& v ) const
    {
      return is_equal( _x , v._x ) && 
	     is_equal( _y , v._y ) &&
             is_equal( _z , v._z ) ;
    }


    /** 
     * \fn bool operator!=( const t3DVector& v ) const
     *
     * \brief Verifies  whether this vector is different  from a given
     * vector.
     *
     * \return  The logic  value false  if  this vector  has the  same
     * coordinates of a given  vector. Otherwise, the logic value true
     * is returned.
     */
    bool operator!=( const t3DVector& v ) const
    {
      return *this == v ;
    }


    /** 
     * \fn t3DVector& operator=( const t3DVector& v )
     *
     * \brief Assigns a given vector to this vector.
     *
     * \param v A vector in 3D space.
     */
    t3DVector& operator=( const t3DVector& v )
    {
      _x = v._x ;
      _y = v._y ;
      _z = v._z ;
      
      return *this ;
    }


    /** 
     * \fn t3DVector& operator+=( const t3DVector& v )
     *
     * \brief Makes this  vector equal to itself plus  a given vector,
     * and then returns this vector.
     *
     * \param v A vector in 3D space.
     *
     * \return This vector after addition.
     */
    t3DVector& operator+=( const t3DVector& v )
    {
      _x += v._x ;
      _y += v._y ;
      _z += v._z ;
      
      return *this ;
    }


    /** 
     * \fn t3DVector& operator-=( const t3DVector& v )
     *
     * \brief Makes this vector equal  to itself minus a given vector,
     * and then returns this vector.
     *
     * \param v A vector in 3D space.
     *
     * \return This vector after subtraction.
     */
    t3DVector& operator-=( const t3DVector& v ) 
    {
      _x -= v._x ;
      _y -= v._y ;
      _z -= v._z ;
      
      return *this;
    }


    /** 
     * \fn t3DVector& operator*=( double alpha )
     *
     * \brief Scales this vector by a given constant factor.
     *
     * \param alpha A scalar.
     *
     * \return This vector after being scaled.
     */
    t3DVector& operator*=( double alpha )
    {
      _x *= alpha ;
      _y *= alpha ;
      _z *= alpha ;
      
      return *this ;
    }


    /** 
     * \fn t3DVector operator-() const
     *
     * \brief Multiplies the coordinates of this vector by -1 and then
     * returns  the  resulting  vector.  This  vector  itself  is  not
     * changed.
     *
     * \return A  vector in 3D  corresponding to minus one  times this
     * vector.
     */
    t3DVector operator-() const
    {
      return t3DVector( -_x , -_y , -_z ) ;
    }


    /** 
     * \fn t3DVector operator+( const t3DVector& v )
     *
     * \brief Returns the vector  resulting from adding this vector to
     * a given vector.
     *
     * \param v A vector in 3D space.
     *
     * \return The vector resulting from adding this vector to a given
     * vector.
     */
    t3DVector operator+( const t3DVector& v ) const
    {
      return t3DVector(
		        _x + v._x , 
		        _y + v._y , 
		        _z + v._z
		      ) ;
    }


    /** 
     * \fn t3DVector operator-( const t3DVector& v ) const
     *
     * \brief Returns  the vector  resulting from subtracting  a given
     * vector from this vector.
     *
     * \param v A vector in 3D space.
     *
     * \return The  vector resulting  subtracting a given  vector from
     * this vector.
     */
    t3DVector operator-( const t3DVector& v ) const
    {
      return t3DVector(
		        _x - v._x , 
		        _y - v._y , 
		        _z - v._z
		      ) ;
    }


    /** 
     * \fn t3DVector operator*( double alpha ) const
     *
     * \brief Returns the vector resulting from scaling this vector by
     * a constant factor.
     *
     * \param alpha A scalar.
     *
     * \return  The vector  resulting from  scaling this  vector  by a
     * constant factor.
     */
    t3DVector operator*( double alpha ) const
    {
      return t3DVector(
		        _x * alpha , 
		        _y * alpha , 
		        _z * alpha
		      ) ;
    }


    /** 
     * \fn friend t3DVector operator*( double alpha, const t3DVector& v )
     *
     * \brief Scales this vector by  a constant factor and returns the
     * vector after scaling.
     *
     * \param alpha A scalar.
     * \param v A vector in 3D space.
     *
     * \return This vector after scaling.
     */
    friend t3DVector operator*( double alpha , const t3DVector& v ) ;


    /** 
     * \fn double operator*( const t3DVector& v ) const
     *
     * \brief  Computes the  dot product  of this  vector and  a given
     * vector.
     *
     * \param v A vector in 3D space.
     *
     * \return The dot product of this vector and a given vector.
     */
    double operator*( const t3DVector& v ) const
    {
      return _x * v._x + _y * v._y + _z * v._z ;
    }


    /** 
     * \fn double operator[]( unsigned i ) const throw ( ExceptionObject )
     *
     * \brief Returns the i-th coordinate of this vetor.
     *
     * \param i The index of a vector coordinate.
     *
     * \return The i-th coordinate of this vector.
     */
    double operator[]( unsigned i ) const throw ( ExceptionObject )
    {
      if ( i > 2 ) {
	std::stringstream ss (
			      std::stringstream::in | 
			      std::stringstream::out
			     ) ;
	ss << "Array index is out of range" ;
	throw ExceptionObject( __FILE__ , __LINE__ , ss.str() ) ;  
      }


      if ( i == 0 ) {
	return _x ;
      }
      else if ( i == 1 ) {
	return _y ;
      }

      return _z ;
    }


    /** 
     * \fn double& operator[]( unsigned i ) const throw ( ExceptionObject )
     *
     * \brief  Returns a  reference  to the  i-th  coordinate of  this
     * vetor.
     *
     * \param i The index of a vector coordinate.
     *
     * \return A reference to the i-th coordinate of this vector.
     */
    double& operator[]( unsigned i ) throw ( ExceptionObject )
    {
      if ( i > 2 ) {
	std::stringstream ss (
			      std::stringstream::in | 
			      std::stringstream::out
			     ) ;
	ss << "Array index is out of range" ;
	throw ExceptionObject( __FILE__ , __LINE__ , ss.str() ) ;  
      }


      if ( i == 0 ) {
	return _x ;
      }
      else if ( i == 1 ) {
	return _y ;
      }

      return _z ;
    }


    /** 
     * \fn double length() const 
     *
     * \brief Returns the length of this vector.
     *
     * \return The length of this vector.
     */
    double length() const 
    {
      return sqrt( sqr( _x ) + sqr( _y ) + sqr( _z ) ) ;
    }


    /** 
     * \fn bool is_null() const 
     *
     * \brief Returns the  logic value true if this  vector has lenght
     * zero, and the logic value false otherwise.
     *
     * \return The  logic value true  if this vector has  lenght zero,
     * and the logic value false otherwise.
     */
    bool is_null() const 
    {
      return is_zero( _x ) && is_zero( _y ) && is_zero( _z ) ;
    }


    /** 
     * \fn t3DVector& negate()
     *
     * \brief  Multiplies the coordinates  of this  vector by  -1, and
     * then  returns  this  vector  after the  coordinate  values  are
     * changed.
     *
     * \return This vector after negating its coordinate values. 
     */
    t3DVector& negate()
    {
      _x = -_x ;
      _y = -_y ;
      _z = -_z ;
      
      return *this ;
    }


    /** 
     * \fn double inner( const t3DVector& v )
     *
     * \brief  Computes the  dot product  of this  vector and  a given
     * vector.
     *
     * \param v A vector in 3D space.
     *
     * \return The dot product of this vector and a given vector.
     */
    double inner( const t3DVector& v ) const
    {
      return this->operator*( v ) ;
    }


    /** 
     * \fn t3DVector& normalize()
     *
     * \brief Normalizes  this vector and then returns  a reference to
     * it.
     *
     * \return A reference to this vector after normalization.
     */
    t3DVector& normalize() 
    {
      double inv_norm = INVERSE( this->length() ) ;

      _x *= inv_norm ;
      _y *= inv_norm ;
      _z *= inv_norm ;

      return *this ;
    }



    /** 
     * \fn t3DVector cross( const t3DVector& v ) const
     *
     * \brief Returns  a 3D vector corresponding to  the cross product
     * of this vector and a given vector.
     *
     * \param v A vector in 3D space.
     *
     * \return A 3D vector corresponding  to the cross product of this
     * vector and a given vector.
     */
    t3DVector cross( const t3DVector& v ) const
    {
      return t3DVector(
		       _y * v._z - _z * v._y ,
		       _z * v._x - _x * v._z ,
		       _x * v._y - _y * v._x
		      ) ;
    }


    // ---------------------------------------------------------------
    //
    // Public attributes
    //
    // ---------------------------------------------------------------

    double _x ;    ///< The first coordinate of the vector.
    double _y ;    ///< The second coordinate of the vector.
    double _z ;    ///< The third coordinate of the vector.

  } ;


  /** 
   * \fn inline t3DVector operator*( double alpha, const t3DVector& v )
   *
   * \brief Scales  this vector by  a constant factor and  returns the
   * vector after scaling.
   *
   * \param alpha A scalar.
   * \param v A vector in 3D space.
   *
   * \return This vector after scaling.
   */
  inline t3DVector operator*( double alpha , const t3DVector& v )
  {
    return v * alpha ;
  }

}         // end of namespace common


/** @} */ //end of group class.

#endif   // t3DVECTOR_HPP
