/** 
 * \file grid.cpp
 *  
 * \brief  This  file contains  the  implementation  of  a class  that
 * represents a 3D cubic grid.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "grid.hpp"        // Grid

#include "mathmacros.hpp"  // is_zero()

#include <sstream>         // std::stringstream


/** 
 * \defgroup MCNameSpace Namespace mc
 * @{
 */

/**
 * \namespace mc
 *
 * \brief  The namespace  mc contains  an adaptation  of  the Marching
 * Cubes algorithm  implementation given  by Jules Bloomenthal  in his
 * web site:
 *
 * http://www.unchainedgeometry.com/jbloom/papers.html
 * 
 */

namespace mc {


  using common::ExceptionObject ;
  using common::t3DVector ;


  // -----------------------------------------------------------------
  //
  // Public methods
  //
  // -----------------------------------------------------------------

  /** 
   * \fn Grid::Grid( unsigned sx , unsigned sy , unsigned sz , const t3DVector& og , const t3DVector& sp ) 
   *
   * \brief Creates an instance of this class.
   *
   * \param sx The number of grid vertices in the X direction.
   * \param sy The number of grid vertices in the Y direction.
   * \param sz The number of grid vertices in the Z direction.
   * \param og The origin point of the grid.
   * \param sp The length of cube edges in each direction.
   */
  Grid::Grid( 
	     unsigned sx ,
	     unsigned sy , 
	     unsigned sz , 
	     const t3DVector& og ,
	     const t3DVector& sp 
	    ) 
  :
    _sizex( sx ) ,
    _sizey( sy ) ,
    _sizez( sz ) ,
    _orig(  og ) ,
    _spac(  sp )
  {
    /*
     * Allocate  memory for  the function  values associated  with the
     * grid vertices.
     */
    _f = ( double*** ) new double**[ _sizex ] ;
    for ( unsigned i = 0 ; i < _sizex ; i++ ) {
      _f[ i ] = ( double** ) new double*[ _sizey ] ;
      for ( unsigned j = 0 ; j < _sizey ; j++ ) {
	_f[ i ][ j ] = ( double* ) new double[ _sizez ] ;
      }
    }   
  }


  /** 
   * \fn Grid::Grid( const Grid& g ) 
   *
   * \brief Creates an instance of this class.
   *
   * \param g A 3D grid to be cloned.
   */
  Grid::Grid( const Grid& g ) 
  :
    _sizex( g._sizex ) ,
    _sizey( g._sizey ) ,
    _sizez( g._sizez ) ,
    _orig(   g._orig ) ,
    _spac(   g._spac )
  {
    /*
     * Allocate  memory for  the function  values associated  with the
     * grid vertices.
     */
    _f = ( double*** ) new double**[ _sizex ] ;

    _f = ( double*** ) new double**[ _sizex ] ;
    for ( unsigned i = 0 ; i < _sizex ; i++ ) {
      _f[ i ] = ( double** ) new double*[ _sizey ] ;
      for ( unsigned j = 0 ; j < _sizey ; j++ ) {
	_f[ i ][ j ] = ( double* ) new double[ _sizez ] ;
      }
    }
 
    for ( unsigned i = 0 ; i < _sizex ; i++ ) {
      for ( unsigned j = 0 ; j < _sizey ; j++ ) {
	for ( unsigned k = 0 ; k < _sizez ; k++ ) {
	  _f[ i ][ j ][ k ] = g._f[ i ][ j ][ k ] ;
	}
      }
    }
  }


  /** 
   * \fn Grid::~Grid() 
   *
   * \brief Destroys an instance of this class.
   */
  Grid::~Grid()
  {
    for ( unsigned i = 0 ; i < _sizex ; i++ ) {
      for ( unsigned j = 0 ; j < _sizey ; j++ ) {
	if ( _f[ i ][ j ] ) {
	  delete[] _f[ i ][ j ] ;
	}
      }
      
      if ( _f[ i ] ) {
	delete[] _f[ i ] ;
      }
    }
  
    _f = 0 ;
  }


  /** 
   * \fn double Grid::get_value( unsigned i , unsigned j , unsigned k ) const throw ( ExceptionObject )
   *
   * \brief Returns the  value of an unknown implicit  function at the
   * grid vertex represented by a given triple of indices, (i,j,k).
   *
   * \return The  value of  an unknown implicit  function at  the grid
   * vertex represented by a given triple of indices, (i,j,k).
   */
  double
  Grid::get_value(
		  unsigned i ,
		  unsigned j ,
		  unsigned k 
		 ) 
    const throw ( ExceptionObject )
  {
    if ( ( i > _sizex ) || ( j > _sizey ) || ( k > _sizez ) ) {
      std::stringstream ss (
			    std::stringstream::in | 
			    std::stringstream::out
			   ) ;
      ss << "Vertex index is out of range" ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str() ) ;
    }
  
    return _f[ i ][ j ][ k ] ;
  }


  /** 
   * \fn void Grid::set_value( unsigned i , unsigned j , unsigned k , double f ) throw ( ExceptionObject )
   *
   * \brief Assigns  a value  of an unknown  implicit function  to the
   * grid vertex represented by a given triple of indices, (i,j,k).
   *
   * \param i The first index of the triple representing the vertex.
   * \param j The second index of the triple representing the vertex.
   * \param k The third index of the triple representing the vertex.
   * \param  f  The value  (at  the  vertex)  of an  unknown  implicit
   * function.
   */
  void
  Grid::set_value(
		  unsigned i ,
		  unsigned j ,
		  unsigned k ,
		  double f 
		 ) 
    throw ( ExceptionObject )
  {
    if ( ( i > _sizex ) || ( j > _sizey ) || ( k > _sizez ) ) {
      std::stringstream ss (
			    std::stringstream::in | 
			    std::stringstream::out
			   ) ;
      ss << "Vertex index is out of range" ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str() ) ;
    }
  
    if ( common::is_zero( f ) ) {
      std::stringstream ss (
			    std::stringstream::in | 
			    std::stringstream::out
			   ) ;
      ss << "Vertex grid cannot be assigned the value 0" ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str() ) ;
    }
  
    _f[ i ][ j ][ k ] = f ;
  }

}         // end of namespace mc

/** @} */ //end of group class.
