/** 
 * \file grid.hpp
 *  
 * \brief This file contains the definition of a class that represents
 * a 3D cubic grid. 
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef GRID_HPP
#define GRID_HPP

#include "exceptionobject.hpp"   // common::ExceptionObject
#include "3dvector.hpp"          // common::t3DVector


/** 
 * \defgroup MCNameSpace Namespace mc
 * @{
 */

/**
 * \namespace mc
 *
 * \brief  The namespace  mc contains  an adaptation  of  the Marching
 * Cubes algorithm  implementation given  by Jules Bloomenthal  in his
 * web site:
 *
 * http://www.unchainedgeometry.com/jbloom/papers.html
 * 
 */

namespace mc {


  using common::ExceptionObject ;
  using common::t3DVector ;


  /**
   * \class Grid
   *
   * \brief This class  represents a 3D cubic grid  in which all cubes
   * have  faces parallel  to the  Cartesian axes,  and have  the same
   * size, i.e., the grid is  a uniform, cubic decomposition of the 3D
   * space.
   */
  class Grid {
  public:
    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /** 
     * \fn Grid( unsigned sx , unsigned sy , unsigned sz , const t3DVector& og , const t3DVector& sp ) 
     *
     * \brief Creates an instance of this class.
     *
     * \param sx The number of grid vertices in the X direction.
     * \param sy The number of grid vertices in the Y direction.
     * \param sz The number of grid vertices in the Z direction.
     * \param og The origin point of the grid.
     * \param sp The length of cube edges in each direction.
     */
    Grid( unsigned sx , unsigned sy , unsigned sz , const t3DVector& og , const t3DVector& sp ) ;


    /** 
     * \fn Grid( const Grid& g ) 
     *
     * \brief Creates an instance of this class.
     *
     * \param g A 3D grid to be cloned.
     */
    Grid( const Grid& g ) ;


    /** 
     * \fn ~Grid() 
     *
     * \brief Destroys an instance of this class.
     */
    ~Grid() ;


    /** 
     * \fn unsigned get_size_x() const
     *
     * \brief Returns the number of grid vertices in the X direction.
     *
     * \return The number of grid vertices in the X direction.
     */
    unsigned get_size_x() const 
    { 
      return _sizex ;
    }


    /** 
     * \fn unsigned get_size_y() const
     *
     * \brief Returns the number of grid vertices in the Y direction.
     *
     * \return The number of grid vertices in the Y direction.
     */
    unsigned get_size_y() const 
    { 
      return _sizey ;
    }


    /** 
     * \fn unsigned get_size_z() const
     *
     * \brief Returns the number of grid vertices in the Z direction.
     *
     * \return The number of grid vertices in the Z direction.
     */
    unsigned get_size_z() const 
    { 
      return _sizez ;
    }


    /** 
     * \fn double get_spacing_x() const
     *
     * \brief Returns the length of the cube edges in the X direction.
     *
     * \return The length of the cube edges in the X direction.
     */
    double get_spacing_x() const 
    { 
      return _spac._x ;
    }


    /** 
     * \fn double get_spacing_y() const
     *
     * \brief Returns the length of the cube edges in the Y direction.
     *
     * \return The length of the cube edges in the Y direction.
     */
    double get_spacing_y() const 
    { 
      return _spac._y ;
    }


    /** 
     * \fn double get_spacing_z() const
     *
     * \brief Returns the length of the cube edges in the Z direction.
     *
     * \return The length of the cube edges in the Z direction.
     */
    double get_spacing_z() const 
    { 
      return _spac._z ;
    }


    /** 
     * \fn t3DVector get_origin() const
     *
     * \brief Returns the origin of this 3D grid.
     *
     * \return The origin of this 3D grid.
     */
    t3DVector get_origin() const 
    { 
      return _orig ;
    }


    /** 
     * \fn double get_value( unsigned i , unsigned j , unsigned k ) const throw (ExceptionObject)
     *
     * \brief Returns the value of an unknown implicit function at the
     * grid vertex represented by a given triple of indices, (i,j,k).
     *
     * \return The value  of an unknown implicit function  at the grid
     * vertex represented by a given triple of indices, (i,j,k).
     */
    double get_value( unsigned i , unsigned j , unsigned k ) const throw (ExceptionObject) ;


    /** 
     * \fn void set_value( unsigned i , unsigned j , unsigned k , double f ) throw (ExceptionObject)
     *
     * \brief Assigns a  value of an unknown implicit  function to the
     * grid vertex represented by a given triple of indices, (i,j,k).
     *
     * \param i The first index of the triple representing the vertex.
     * \param  j  The second  index  of  the  triple representing  the
     * vertex.
     * \param k The third index of the triple representing the vertex.
     * \param  f The  value (at  the  vertex) of  an unknown  implicit
     * function.
     */
    void set_value( unsigned i , unsigned j , unsigned k , double f ) throw (ExceptionObject) ;

  
  private:
    // ---------------------------------------------------------------
    //
    // Private attributes
    //
    // ---------------------------------------------------------------

    unsigned _sizex ;    ///< The number of grid vertices in the X direction.
    unsigned _sizey ;    ///< The number of grid vertices in the Y direction.
    unsigned _sizez ;    ///< The number of grid vertices in the Z direction.
    t3DVector _orig ;    ///< The origin of the 3D grid.
    t3DVector _spac ;    ///< The length of the cube edges in each direction.
    double ***_f ;       ///< The values of an unknown implicit function at the vertices of the 3D grid.
  } ;

}         // end of namespace mc


/** @} */ //end of group class.

#endif   // GRID_HPP
