/** 
 * \file surfbuilder.cpp
 *  
 * \brief This  file contains the  implementation of the methods  of a
 * class for  computing a  simplicial approximation for  an implicitly
 * defined surface using the Marching Cubes algorithm.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "surfbuilder.hpp"    // mc::SurfBuilder

#include "mathmacros.hpp"     // SIGN

#include <sstream>            // std::stringstream


/** 
 * \defgroup MCNameSpace Namespace mc
 * @{
 */

/**
 * \namespace mc
 *
 * \brief  The namespace  mc contains  an adaptation  of  the Marching
 * Cubes algorithm  implementation given  by Jules Bloomenthal  in his
 * web site:
 *
 * http://www.unchainedgeometry.com/jbloom/papers.html
 * 
 */

namespace mc {


  using common::t3DVector ;
  using common::ExceptionObject ;


  // -----------------------------------------------------------------
  //
  // Public methods
  //
  // -----------------------------------------------------------------

  /** 
   * \fn void SurfBuilder::run() 
   *
   * \brief  Creates the  simplicial surface  (i.e., a  triangle mesh)
   * that  approximates the  surface defined  by the  unknown implicit
   * function.
   */
  void 
  SurfBuilder::run() 
  {
    /*
     * Create a  table to  store all possible  ways of  traversing the
     * edges of a grid cube  to create the triangles of the simplicial
     * surface.
     */
    make_lookup_table() ;

    /*
     * For each cube belonging to the grid and containing the vertices
     * (i,j,k),   (i+1,j,k),    (i,j+1,k),   (i+1,j+1,k),   (i,j,k+1),
     * (i+1,j,k+1), (i,j+1,k+1),  (i+1,j+1,k+1), obtain the  values of
     * the unknown implicit function at  the vertices of the cube, and
     * then compute the intersection  points, if any, between the cube
     * edges and the surface defined by the unknown implicit function.
     */

    /*
     * Get the number of grid points in each dimension of the grid. 
     */
    unsigned sx = _grid->get_size_x() - 1 ;
    unsigned sy = _grid->get_size_y() - 1 ;
    unsigned sz = _grid->get_size_z() - 1 ;

    /*
     * Create an array to store the function values.
     */
    double fv[ 8 ] ;

    /*
     * Visit each cube of the grid and compute the intersection points
     * between  its  edges and  the  surface  defined  by the  unknown
     * implicit function.
     */
    for ( unsigned k = 0 ; k < sz ; k++ ) {
      for ( unsigned j = 0 ; j < sy ; j++ ) {
	for ( unsigned i = 0 ; i < sx ; i++ ) {
	  /*
	   * Get the value of the implicit function at the vertices of
	   * the cube whose origin vertex is given by the triple index
	   * (i,j,k).
	   */
	  fv[ 0 ] = _grid->get_value(     i ,     j ,     k ) ;
	  fv[ 1 ] = _grid->get_value(     i ,     j , k + 1 ) ;
	  fv[ 2 ] = _grid->get_value(     i , j + 1 ,     k ) ;
	  fv[ 3 ] = _grid->get_value(     i , j + 1 , k + 1 ) ;
	  fv[ 4 ] = _grid->get_value( i + 1 ,     j ,     k ) ;
	  fv[ 5 ] = _grid->get_value( i + 1 ,     j , k + 1 ) ;
	  fv[ 6 ] = _grid->get_value( i + 1 , j + 1 ,     k ) ;
	  fv[ 7 ] = _grid->get_value( i + 1 , j + 1 , k + 1 ) ;

	  /*
	   * Determine an entry of the  lookup table by using the sign
           * of the values of the implicit function at the vertices of
           * the cube.
	   */
	  unsigned n , index = 0 ;
	  for ( n = 0 ; n < 8 ; n++ ) {
	    if ( fv[ n ] > 0 ) {
	      index += unsigned( 1 << n ) ;
	    }
	  }

	  /*
	   * Traverse the faces of the cube according to the traversal
	   * stored in  the lookup table  at the table  entry computed
	   * above.
	   */
	  for (
	       TABLEENTRY::const_iterator tit = _cubetable[ index ].begin() ;
	       tit != _cubetable[ index ].end() ;
	       ++tit
	      ) 
	  {
	    /*
	     * Get the  edges intersected  by the implicit  surface in
	     * the order  they appear in the traversals  stored in the
	     * table.
	     */
	    for (
		 TRAVERSAL::const_iterator eit = (*tit).begin() ;
		 eit != (*tit).end() ;
		 eit++
		 ) 
	    {
	      /*
	       * Get the edge index.
	       */
	      EDGES ie = (EDGES) *eit ;

	      /*
	       * Get the indices of the two vertices incident to the edge.
	       */
	      VERTS v1 = get_1st_vertex( ie ) ;
	      VERTS v2 = get_2nd_vertex( ie ) ;

	      /*
	       * Compute the  intersection point between  the edge and
	       * the surface defined by the unknown implicit function.
	       */
	      t3DVector pt ;
	      try {
		compute_intersection(
				     i ,
				     j ,
				     k ,
				     v1 ,
				     v2 , 
				     fv[ v1 ] ,
				     fv[ v2 ] ,
				     pt
				    ) ;
	      }
	      catch ( const ExceptionObject& xpt ) {
		treat_exception( xpt ) ;
		exit( 0 ) ;
	      }

	      // DO SOMETHING WITH THE POINT JUST COMPUTED
	    }

	    // DO SOMETHING WITH ALL POINTS OF INTERSECTION (THEY DEFINE ONE PATCH)
	  }
	}
      }
    }

    return ;
  }


  /** 
   * \fn void SurfBuilder::make_lookup_table()
   *
   * \brief Builds a  table with 256 entries. Each  entry is related
   * to a possible (distinct) way  of assigning one of two values to
   * the  vertices  of a  cube.  Each entry  also  stores  a set  of
   * traversals,  each  of which  contains  the  set  of cube  edges
   * intersected  by the  surface  defined by  the unknown  implicit
   * function.
   */
  void
  SurfBuilder::make_lookup_table()
  {
    /*
     * For each  possible way of assigning  one of two  values for the
     * eight vertices  of a cube,  store traversal information  into a
     * table entry.  Each traversal contains a sequence  of cube edges
     * intersected  by the  surface  defined by  the unknown  implicit
     * function.
     */
    for ( unsigned i = 0 ; i < 256 ; i++ ) {
      /*
       * Initialize an array to indicate  whether a grid cube edge has
       * been visited during one traversal.
       */
      bool done[ 12 ] ;
      for ( unsigned e = 0 ; e < 12 ; e++ ) {
	done[ e ] = false ;
      }

      /*
       * Compute  an  array with  powers  of  2  corresponding to  the
       * indices of the vertices of  the same cube. The vertex indices
       * are 0, 1, ..., 7, and the respective power of 2 are 2^0, 2^1,
       * ..., 2^7.
       */
      bool pos[ 8 ] ;
      for ( unsigned j = 0 ; j < 8 ; j++ ) {
	pos[ j ] = compute_sign( i , j ) ;
      }

      /*
       * Compute a traversal  of the cube that determines  the way the
       * points inside the cube must  be connected to form one or more
       * triangles.
       */
      for ( unsigned e = 0 ; e < 12 ; e++ ) {
	/*
	 * For each edge of the cube that has not been visited yet and
	 * whose  vertices   have  opposite  signs,   initiate  a  new
	 * traversal  and insert  the  traversal into  a lookup  table
	 * entry.
	 */
	EDGES ie = ( EDGES ) e ;

	if ( !done[ ie ] && 
	     ( pos[ get_1st_vertex( ie ) ] != pos[ get_2nd_vertex( ie ) ] ) 
	   ) 
	{
	  EDGES stae = ie ;
	  EDGES edge = ie ;

	  /*
	   * Obtain the cube face on  the right side of the edge "ie",
	   * where  "side"  is  defined   with  respect  to  the  edge
	   * orientation given  from the vertex with  positive sign to
	   * the vertex with negative sign.
	   */
	  FACES face = pos[ get_1st_vertex( ie ) ] ? 
	               right_face( ie ) : left_face( ie ) ;

	  /*
	   * Create a traversal.
	   */
	  TRAVERSAL curr ;

	  /*
	   * Compute  a traversal along  the faces  of the  cube. This
	   * traversal  consists  of  edges  whose  the  sign  of  the
	   * implicit  function  at  the  two  incident  vertices  are
	   * opposite.
	   */
	  do {
	    /*
	     * Pick the next edge in the current face.
	     */
	    edge = next_cw_edge( edge , face ) ;

	    /*
	     * Mark the new current edge as visited.
	     */
	    done[ edge ] = true ;

	    /*
	     * If the sign of the implicit function at the vertices of
	     * the current  edge are opposite, insert the  edge in the
	     * current traversal and change face to find the remaining
	     * edges.
	     */
	    if ( pos[ get_1st_vertex( edge ) ] != pos[ get_2nd_vertex( edge ) ] ) {
	      curr.push_back( edge ) ;
	      face = change_face(edge , face ) ;
	    }

	  } while ( stae != edge ) ;

	  /*
	   * Insert the new traversal into the lookup table.
	   */
	  _cubetable[ i ].push_back( curr ) ;
	}
      }
    }

    return ;
  }


  /** 
   * \fn SurfBuilder::VERTS SurfBuilder::get_1st_vertex( EDGES e ) const
   *
   * \brief Returns the index of the first vertex of a given edge.
   *
   * \param e The index of an edge of a grid cube.
   *
   * \return The index of the first vertex of a given edge.
   */
  SurfBuilder::VERTS 
  SurfBuilder::get_1st_vertex( EDGES e ) const
  {
    VERTS index ;

    if ( e == LB ) {
      index = LBN ;
    }
    else if ( e == LT ) {
      index = LTN ;
    }
    else if ( e == LN ) {
      index = LBN ;
    }
    else if ( e == LF ) {
      index = LBF ;
    }
    else if ( e == RB ) {
      index = RBN ;
    }
    else if ( e == RT ) {
      index = RTN ;
    }
    else if ( e == RN ) {
      index = RBN ;
    }
    else if ( e == RF ) {
      index = RBF ;
    }
    else if ( e == BN ) {
      index = LBN ;
    }
    else if ( e == BF ) {
      index = LBF ;
    }
    else if ( e == TN ) {
      index = LTN ;
    }
    else { // if ( e == TF ) {
      index = LTF ;
    }

    return index ;
  }


  /** 
   * \fn SurfBuilder::VERTS SurfBuilder::get_2nd_vertex( EDGES e ) const
   *
   * \brief Returns the index of the second vertex of a given edge.
   *
   * \param e The index of an edge of a grid cube.
   *
   * \return The index of the second vertex of a given edge.
   */
  SurfBuilder::VERTS
  SurfBuilder::get_2nd_vertex( EDGES e ) const
  {
    VERTS index ;

    if ( e == LB ) {
      index = LBF ;
    }
    else if ( e == LT ) {
      index = LTF ;
    }
    else if ( e == LN ) {
      index = LTN ;
    }
    else if ( e == LF ) {
      index = LTF ;
    }
    else if ( e == RB ) {
      index = RBF ;
    }
    else if ( e == RT ) {
      index = RTF ;
    }
    else if ( e == RN ) {
      index = RTN ;
    }
    else if ( e == RF ) {
      index = RTF ;
    }
    else if ( e == BN ) {
      index = RBN ;
    }
    else if ( e == BF ) {
      index = RBF ;
    }
    else if ( e == TN ) {
      index = RTN ;
    }
    else { // if ( e == TF ) {
      index = RTF ;
    }

    return index ;
  }


  /** 
   * \fn SurfBuilder::FACES SurfBuilder::left_face( EDGES e ) const
   *
   * \brief  Returns the  index  of the  left  face of  a given  edge,
   * considering the  edge from the  vertex with positive sign  to the
   * vertex with negative sign.
   *
   * \param e The index of an edge of a grid cube.
   *
   * \return The index of the left face of a given edge.
   */
  SurfBuilder::FACES
  SurfBuilder::left_face( EDGES e ) const
  {
    FACES index ;

    if ( e == LB ) {
      index = B ;
    }
    else if ( e == LT ) {
      index = L ;
    }
    else if ( e == LN ) {
      index = L ;
    }
    else if ( e == LF ) {
      index = F ;
    }
    else if ( e == RB ) {
      index = R ;
    }
    else if ( e == RT ) {
      index = T ;
    }
    else if ( e == RN ) {
      index = N ;
    }
    else if ( e == RF ) {
      index = R ;
    }
    else if ( e == BN ) {
      index = N ;
    }
    else if ( e == BF ) {
      index = B ;
    }
    else if ( e == TN ) {
      index = T ;
    }
    else { // if ( e == TF ) {
      index = F ;
    }

    return index ;
  }


  /** 
   * \fn SurfBuilder::FACES SurfBuilder::right_face( EDGES e ) const
   *
   * \brief  Returns the  index of  the right  face of  a  given edge,
   * considering the  edge from the  vertex with positive sign  to the
   * vertex with negative sign.
   *
   * \param e The index of an edge of a grid cube.
   *
   * \return The index of the right face of a given edge.
   */
  SurfBuilder::FACES 
  SurfBuilder::right_face( EDGES e ) const
  {
    FACES index ;

    if ( e == LB ) {
      index = L ;
    }
    else if ( e == LT ) {
      index = T ;
    }
    else if ( e == LN ) {
      index = N ;
    }
    else if ( e == LF ) {
      index = L ;
    }
    else if ( e == RB ) {
      index = B ;
    }
    else if ( e == RT ) {
      index = R ;
    }
    else if ( e == RN ) {
      index = R ;
    }
    else if ( e == RF ) {
      index = F ;
    }
    else if ( e == BN ) {
      index = B ;
    }
    else if ( e == BF ) {
      index = F ;
    }
    else if ( e == TN ) {
      index = N ;
    }
    else { // if ( e == TF ) {
      index = T ;
    }

    return index ;
  }


  /** 
   * \fn SurfBuilder::EDGES SurfBuilder::next_cw_edge( EDGES e , FACES f ) const
   *
   * \brief Returns the index of the edge that follows a given edge in
   * a clockwise  traversal of the edges  of a given face  of the grid
   * cube.
   *
   * \param e The index of an edge of a grid1 cube.
   * \param f  The index  of a grid  cube face containing  the given
   * edge.
   *
   * \return The  index of the edge  that follows a given  edge in a
   * clockwise traversal  of the edges of  a given face  of the grid
   * cube.
   */
  SurfBuilder::EDGES
  SurfBuilder::next_cw_edge( EDGES e , FACES f ) const
  {
    EDGES index ;

    if ( e == LB ) {
      index = ( f == L ) ? LF : BN ;
    }
    else if ( e == LT ) {
      index = ( f == L ) ? LN : TF ;
    }
    else if ( e == LN ) {
      index = ( f == L ) ? LB : TN ;
    }
    else if ( e == LF ) {
      index = ( f == L ) ? LT : BF ;
    }
    else if ( e == RB ) {
      index = ( f == R ) ? RN : BF ;
    }
    else if ( e == RT ) {
      index = ( f == R ) ? RF : TN ;
    }
    else if ( e == RN ) {
      index = ( f == R ) ? RT : BN ;
    }
    else if ( e == RF ) {
      index = ( f == R ) ? RB : TF ;
    }
    else if ( e == BN ) {
      index = ( f == B ) ? RB : LN ;
    }
    else if ( e == BF ) {
      index = ( f == B ) ? LB : RF ;
    }
    else if ( e == TN ) {
      index = ( f == T ) ? LT : RN ;
    }
    else { // if ( e == TF ) {
      index = ( f == T ) ? RT : LF ;
    }

    return index ;
  }


  /** 
   * \fn void SurfBuilder::compute_intersection( unsigned i , unsigned j , unsigned k , VERTS v1 , VERTS v2 , double f1 , double f2 , t3DVector& pt ) const throw( ExceptionObject )
   *
   * \brief Computes the coordinates of the intersection point between
   * a given  edge and the  surface defined by the  (unknown) implicit
   * function.
   *
   * \param i  First element  of a triple  that identifies  the origin
   * vertex of the grid cube containing the edge.
   * \param j  Second element of  a triple that identifies  the origin
   * vertex of the grid cube containing the edge.
   * \param k  Third element  of a triple  that identifies  the origin
   * vertex of the grid cube containing the edge.
   * \param v1 The index of the  first vertex of the edge with respect
   * to the grid cube.
   * \param v2 The index of the second vertex of the edge with respect
   * to the grid cube.
   * \param f1 The value of the unknown implicit function at the first
   * vertex of the edge.
   * \param  f2 The  value of  the  unknown implicit  function at  the
   * second vertex of the edge.
   * \param pt A reference to the intersection point.
   */
  void
  SurfBuilder::compute_intersection(
				    unsigned i ,
				    unsigned j ,
				    unsigned k ,
				    VERTS v1 ,
				    VERTS v2 ,
				    double f1 ,
				    double f2 ,
				    t3DVector& pt
				   )
    const throw( ExceptionObject )
  {
    if ( SIGN( f1 ) == SIGN( f2 ) ) {
      std::stringstream ss ( std::stringstream::in | 
			     std::stringstream::out ) ;
      ss << "Trying to compute intersection point on non-interesecting edge" ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str() ) ;
    }

    unsigned i1 , j1 , k1 , i2 , j2 , k2 ;

    /*
     * Initialize indices of "v1" and "v2".
     */
    i1 = i2 = i ;
    j1 = j2 = j ;
    k1 = k2 = k ;

    /*
     * Compute the indices of "v1" and "v2" with respect to the grid.
     */
    if ( ( v1 == RBN ) || ( v1 == RBF ) || ( v1 == RTN ) || ( v1 == RTF ) ) {
      i1 += 1 ;
    }

    if ( ( v1 == LTN ) || ( v1 == LTF ) || ( v1 == RTN ) || ( v1 == RTF ) ) {
      j1 += 1 ;
    }

    if ( ( v1 == LBF ) || ( v1 == LTF ) || ( v1 == RBF ) || ( v1 == RTF ) ) {
      k1 += 1 ;
    }

    if ( ( v2 == RBN ) || ( v2 == RBF ) || ( v2 == RTN ) || ( v2 == RTF ) ) {
      i2 += 1 ;
    }

    if ( ( v2 == LTN ) || ( v2 == LTF ) || ( v2 == RTN ) || ( v2 == RTF ) ) {
      j2 += 1 ;
    }

    if ( ( v2 == LBF ) || ( v2 == LTF ) || ( v2 == RBF ) || ( v2 == RTF ) ) {
      k2 += 1 ;
    }

    /*
     * DO SOMETHING TO AVOID COMPUTING THE SAME VERTEX TWICE!
     */

    /*
     * Obtain the coordinates of vertices "v1" and "v2".
     */
    t3DVector go = _grid->get_origin() ;

    double spacx = _grid->get_spacing_x() ;
    double spacy = _grid->get_spacing_y() ;
    double spacz = _grid->get_spacing_z() ;
 
    t3DVector p1 = go + t3DVector( spacx * i1 , spacy * j1 , spacz * k1 ) ;
    t3DVector p2 = go + t3DVector( spacx * i2 , spacy * j2 , spacz * k2 ) ;

    /*
     * Compute position  of intersection  point bewteen the  cube edge
     * and the surface defined  by the unknown implicit function using
     * affine interpolation.
     */
    const double coeff = f2 / ( f2 - f1 ) ;
    pt = ( coeff * p1 ) + ( ( 1 - coeff ) * p2 ) ;

    // DO SOMETHING WITH THE INTERSECTION POINT

    return ;
  }

}         // end of namespace mc

/** @} */ //end of group class.
