/** 
 * \file surfbuilder.hpp
 *  
 * \brief This file contains the definition of a class for computing a
 * simplicial  approximation for an  implicitly defined  surface using
 * the Marching Cubes algorithm.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef SURFBUILDER_HPP
#define SURFBUILDER_HPP

#include <list>                // std::list
#include <vector>              // std::vector

#include "exceptionobject.hpp" // commonlib::ExceptionObject
#include "3dvector.hpp"        // commonlib::t3DVector
#include "grid.hpp"            // mc::grid


/** 
 * \defgroup MCNameSpace Namespace mc
 * @{
 */

/**
 * \namespace mc
 *
 * \brief  The namespace  mc contains  an adaptation  of  the Marching
 * Cubes algorithm  implementation given  by Jules Bloomenthal  in his
 * web site:
 *
 * http://www.unchainedgeometry.com/jbloom/papers.html
 * 
 */

namespace mc {


  using common::t3DVector ;
  using common::ExceptionObject ;


  /**
   * \class SurfBuilder
   *
   * \brief This  class implements an improved version  of the orginal
   * Marching Cubes  algorithm. The algorithm  prevents the generation
   * of polygonal surfaces with  holes. However, it does not guarantee
   * that the polygonal surface has  the same topology as the implicit
   * surface defined  by the  function that affinely  interpolates the
   * values of the  unknown function in the grid  cubes of the spatial
   * decomposition.
   */
  class SurfBuilder {
  public:
    // ---------------------------------------------------------------
    //
    // Type definitions
    //
    // ---------------------------------------------------------------

    /**
     * \enum FACES
     *
     * \brief Defines an index for each face of a grid cube.
     */    
    enum FACES { L , R , B , T , N , F } ;


    /**
     * \enum EDGES
     *
     * \brief Defines an index for each edge of a grid cube.
     */    
    enum EDGES { 
                 LB , LT , LN , LF ,
                 RB , RT , RN , RF , 
                 BN , BF , TN , TF 
               } ;


    /**
     * \enum VERTS
     *
     * \brief Defines an index for each vertex of a grid cube.
     */    
    enum VERTS { LBN , LBF , LTN , LTF , RBN , RBF , RTN , RTF } ;


    /**
     * \typedef TRAVERSAL
     *
     * \brief Defines a type for  a list containing the indices of the
     * edges of a cube that  are intersected by the (unknown) implicit
     * function.
     */    
    typedef std::list< unsigned >  TRAVERSAL ;


    /**
     * \typedef TABLEENTRY
     *
     * \brief Defines a type for  a list of traversals, such that each
     * traversal is  itself a  list of  the edges of  a cube  that are
     * intersected by the (unknown) implicit function.
     */
    typedef std::list< TRAVERSAL > TABLEENTRY ;


    /**
     * \typedef LOOKUPTABLE
     *
     * \brief Defines  a table for  storing all traversals  related to
     * each  one  of  256  possible  ways of  assigning  positive  and
     * negative values  to an ordered  sequence of vertices of  a grid
     * cube.
     */
    typedef std::vector< TABLEENTRY > LOOKUPTABLE ;


    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /** 
     * \fn SurfBuilder( Grid* ) 
     *
     * \brief Creates an instance of this class.
     *
     * \param g A pointer to a  3D grid with the vertices of all cubes
     * of a spatial decomposition along with the values of the unknown
     * implicit function at the vertices.
     */
    SurfBuilder( Grid* g ) : _grid( g )
    {
      _cubetable.resize( 256 ) ;
    }


    /** 
     * \fn ~SurfBuilder() 
     *
     * \brief Destroys an instance of this class.
     */
    ~SurfBuilder()
    {}


    /** 
     * \fn void run() 
     *
     * \brief Creates  the simplicial surface (i.e.,  a triangle mesh)
     * that approximates  the surface defined by  the unknown implicit
     * function.
     */    
    void run() ;


  private:
    // ---------------------------------------------------------------
    //
    // Private methods
    //
    // ---------------------------------------------------------------

    /** 
     * \fn void make_lookup_table()
     *
     * \brief Builds a  table with 256 entries. Each  entry is related
     * to a possible (distinct) way  of assigning one of two values to
     * the  vertices  of a  cube.  Each entry  also  stores  a set  of
     * traversals,  each  of which  contains  the  set  of cube  edges
     * intersected  by the  surface  defined by  the unknown  implicit
     * function.
     */
    void make_lookup_table() ;


    /** 
     * \fn bool compute_sign( unsigned n , unsigned i ) const
     *
     * \brief Returns  true if  and only  if the i-th  bit of  a given
     * number is 1.
     *
     * \param n A given number.
     * \param i Another given number representing the position of bit.
     *
     * \return The logic value true if  the i-th bit of a given number
     * is 1. Otherwise, returns the logic value false.
     */
    bool compute_sign(
		      unsigned n ,
		      unsigned i
		     )
      const
    {
      return  ( ( ( n ) >> ( i ) ) & 1 ) == 1 ;
    }


    /** 
     * \fn VERTS get_1st_vertex( EDGES e ) const
     *
     * \brief Returns the index of the first vertex of a given edge.
     *
     * \param e The index of an edge of a grid cube.
     *
     * \return The index of the first vertex of a given edge.
     */
    VERTS get_1st_vertex( EDGES e ) const ;


    /** 
     * \fn VERTS get_2nd_vertex( EDGES e ) const
     *
     * \brief Returns the index of the second vertex of a given edge.
     *
     * \param e The index of an edge of a grid cube.
     *
     * \return The index of the second vertex of a given edge.
     */
    VERTS get_2nd_vertex( EDGES e ) const ;


    /** 
     * \fn FACES left_face( EDGES e ) const
     *
     * \brief  Returns the index  of the  left face  of a  given edge,
     * considering the edge from the  vertex with positive sign to the
     * vertex with negative sign.
     *
     * \param e The index of an edge of a grid cube.
     *
     * \return The index of the left face of a given edge.
     */
    FACES left_face( EDGES e ) const ;


    /** 
     * \fn FACES right_face( EDGES e ) const
     *
     * \brief Returns  the index  of the right  face of a  given edge,
     * considering the edge from the  vertex with positive sign to the
     * vertex with negative sign.
     *
     * \param e The index of an edge of a grid cube.
     *
     * \return The index of the right face of a given edge.
     */
    FACES right_face( EDGES e ) const ;


    /** 
     * \fn EDGES next_cw_edge( EDGES e , FACES f ) const
     *
     * \brief Returns the index of  the edge that follows a given edge
     * in a  clockwise traversal of the  edges of a given  face of the
     * grid cube.
     *
     * \param e The index of an edge of a grid1 cube.
     * \param f  The index  of a grid  cube face containing  the given
     * edge.
     *
     * \return The  index of the edge  that follows a given  edge in a
     * clockwise traversal  of the edges of  a given face  of the grid
     * cube.
     */
    EDGES next_cw_edge( EDGES e , FACES f ) const ;


    /** 
     * \fn FACES change_face( EDGES e , FACES f ) const
     *
     * \brief Returns the index of a face incident to an edge.
     *
     * \param e The index of an edge of a grid cube.
     * \param f  The index of a  grid cube face incident  to the given
     * edge, which  is not  equal to the  other face that  this method
     * should return.
     *
     * \return The index of a face incident to an edge.
     */
    FACES change_face( EDGES e , FACES f ) const
    {
      FACES index = left_face( e ) ;

      return ( f == index ) ? right_face( e ) : index ;
    }


    /** 
     * \fn unsigned from_index_to_id( unsigned i , unsigned j , unsigned k ) const
     *
     * \brief  Returns the  unique identifier  of a  grid  cube vertex
     * corresponding to  a given  triple (  i , j  , k)  that uniquely
     * identifies the vertex in the grid.
     *
     * \param i First element of the vertex index.
     * \param j Second element of the vertex index.
     * \param k Third element of the vertex index.
     *
     * \return   The  unique   identifier  of   a  grid   cube  vertex
     * corresponding to  a given  triple (  i , j  , k)  that uniquely
     * identifies the vertex in the grid.
     */
    unsigned from_index_to_id(
			      unsigned i ,
			      unsigned j ,
			      unsigned k
			     )
      const
    {
      unsigned sx = _grid->get_size_x() ;
      unsigned sy = _grid->get_size_y() ;

      return i + ( j * sx ) + ( k * sx * sy ) ;
    }


    /** 
     * \fn void compute_intersection( unsigned i , unsigned j , unsigned k , VERTS v1 , VERTS v2 , double f1 , double f2 , t3DVector& pt ) const throw( ExceptionObject )
     *
     * \brief  Computes  the  coordinates  of the  intersection  point
     * between a given  edge and the surface defined  by the (unknown)
     * implicit function.
     *
     * \param i First  element of a triple that  identifies the origin
     * vertex of the grid cube containing the edge.
     * \param j Second element of  a triple that identifies the origin
     * vertex of the grid cube containing the edge.
     * \param k Third  element of a triple that  identifies the origin
     * vertex of the grid cube containing the edge.
     * \param  v1 The  index  of the  first  vertex of  the edge  with
     * respect to the grid cube.
     * \param  v2 The  index of  the second  vertex of  the  edge with
     * respect to the grid cube.
     * \param f1  The value  of the unknown  implicit function  at the
     * first vertex of the edge.
     * \param f2  The value  of the unknown  implicit function  at the
     * second vertex of the edge.
     * \param pt A reference to the intersection point.
     */
    void compute_intersection( unsigned i , unsigned j , unsigned k ,
      VERTS v1 , VERTS v2 , double f1 , double f2 , t3DVector& pt ) 
      const throw( ExceptionObject ) ;

    // ---------------------------------------------------------------
    //
    // Private attributes
    //
    // ---------------------------------------------------------------

    Grid* _grid ;               ///< A 3D grid with function values associated with the grid vertices.
    LOOKUPTABLE _cubetable ;    ///< A table that stores the traversals of the grid cubes.

  } ;


}         // end of namespace mc


/** @} */ //end of group class.

#endif   // SURFBUILDER_HPP
