/** 
 * \file tst-stdmc.cpp
 *  
 * \brief A driver to test the implementation of the MC algorithm.
 *
 * \author
 * Marcelo Ferreira Siqueira
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * mfsiqueira at gmail (dot) com
 *
 * \version 1.0
 * \date August 2012
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include <iostream>           // std::cout, std::endl, std::cerr
#include <cstdlib>            // EXIT_SUCCESS, EXIT_FAILURE
#include <fstream>            // std::ifstream, std::ofstream

#include "grid.hpp"           // mc::Grid
#include "surfbuilder.hpp"    // mc::SurfBuilder
#include "3dvector.hpp"       // common::t3DVector


using mc::Grid ;
using mc::SurfBuilder ;
using common::t3DVector ;


// -------------------------------------------------------------------
//
// Auxiliary functions.
//
// -------------------------------------------------------------------

/**
 * \fn bool check_grid( const Grid& g ) 
 *
 * \brief Decides whether a given 3D grid is consistent, i.e., whether
 * all values assigned  to the vertices of the  grid are distinct from
 * zero and  the values  assigned to the  boundary points  are greater
 * than zero.
 *
 * \param g A 3D grid
 *
 * \return The  logic value  true of the  grid is consistent,  and the
 * logic value false otherwise.
 */
bool check_grid( const Grid& g ) 
{
  unsigned sx = g.get_size_x() ;
  unsigned sy = g.get_size_y() ;
  unsigned sz = g.get_size_z() ;

  for ( unsigned j = 0 ; j < sy ; j++ ) {
    for ( unsigned i = 0 ; i < sx ; i++ ) {
      if ( 
	  ( g.get_value( i , j ,      0 ) <= 0 ) ||
	  ( g.get_value( i , j , sz - 1 ) <= 0 ) 
	 ) 
      {
        return false ;
      }
    }
  }

  for ( unsigned k = 0 ; k < sz ; k++ ) {
    for ( unsigned i = 0 ; i < sx ; i++ ) {
      if ( 
	  ( g.get_value( i ,      0 , k ) <= 0 ) || 
	  ( g.get_value( i , sy - 1 , k ) <= 0 ) 
	 ) 
      {
        return false ;
      }
    }
  }

  for ( unsigned k = 0 ; k < sz ; k++ ) {
    for ( unsigned j = 0 ; j < sy ; j++ ) {
      if ( 
	  ( g.get_value(     0 , j , k ) <= 0 ) || 
	  ( g.get_value( sx - 1, j , k ) <= 0 )
	 ) 
      {
        return false ;
      }
    }
  }

  return true ;
}


// ------------------------------------------------------------------
//
// Main function starts below:
//
// ------------------------------------------------------------------

/**
 * \fn int main()
 *
 * \brief A simple test for an implementation of the MC algorithm.
 *
 * \param argc Number of command-line parameters.
 * \param argv An array with the command-line parameters.
 *
 * \return An integer number.
 */
int main( int argc , char* argv[] ) 
{
  if ( argc != 3 ) {
    std::cerr << std::endl ;
    std::cerr << "\t Usage: tst-stdmc < arg1 > < arg2 >" 
              << std::endl ;
    std::cerr << "\t\t arg1   the input filename" 
	      << std::endl ;
    std::cerr << "\t\t arg2   the output filename" 
	      << std::endl ;
    std::cerr << std::endl ;

    return EXIT_FAILURE ;
  }

  /*
   * Abre o arquivo de entrada.
   */
  std::ifstream pf1( argv[ 1 ] ) ;

  if ( pf1.is_open() ) {
    //
    // Read in the grid data.
    //

    /*
     * Read in the number of grid points in each direction.
     */
    unsigned sizex , sizey , sizez ;
    pf1 >> sizex >> sizey >> sizez ;

    /*
     * Read in the spacing between the grid points.
     */
    double spacx , spacy , spacz ;
    pf1 >> spacx >> spacy >> spacz ;
    
    /*
     * Read in the grid origin.
     */
    double origx , origy , origz ;
    pf1 >> origx >> origy >> origz ;

    std::cerr << "Create a 3D grid..." << std::endl ;

    Grid g( 
	   sizex ,                               // Number of points in the X direction.
	   sizey ,                               // Number of points in the Y direction.
	   sizez ,                               // Number of points in the Z direction.
	   t3DVector( origx , origy , origz ) ,  // Grid origin.
	   t3DVector( spacx , spacy , spacz )    // Length of the edges in each direction.
	  ) ;

    /*
     * Read in the values associated with the grid points.
     */
    for ( unsigned k = 0 ; k < sizez ; k++ ) {
      for ( unsigned j = 0 ; j < sizey ; j++ ) {
	for ( unsigned i = 0 ; i < sizex ; i++ ) {
	  double f ;
	  pf1 >> f ;
	  g.set_value( i , j , k , f ) ;
	}
      }
    }

    /*
     * Close the input file.
     */
    pf1.close() ;

    //
    // Check grid consistency.
    //
    
    std::cerr << "Check the consistency of the grid..." 
	      << std::endl ;
    
    if ( !check_grid( g ) ) {
      std::cerr << "The grid is not consistent" << std::endl ;
      return EXIT_FAILURE;
    }

    //
    // Run the MC algorithm.
    //

    std::cerr << "Run the Marching Cubes algorithm..." 
	      << std::endl ;

    SurfBuilder sb( &g ) ;
    
    sb.run() ;
    
    std::cerr << "Done!" << std::endl ;

    //
    // Create the output file and write out the mesh data there
    //
    std::ofstream pf2( argv[ 2 ] ) ;

    if ( pf2.is_open() ) {
      //
      // Write out the mesh data
      //

      pf2 << "No mesh data!" << std::endl ;

      pf2.close() ;
    }
    else {
      std::cerr << "Can't open the output file!" << std::endl ;
      return EXIT_FAILURE ;
    }
  }
  else {
    std::cerr << "Can't open the input file!" << std::endl ;
    return EXIT_FAILURE ;
  }
    
  return EXIT_SUCCESS ;
}



